#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Any
from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.language_v1 import Document

from airflow.providers.google.cloud.hooks.natural_language import CloudNaturalLanguageHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_no_default_project_id

API_RESPONSE: dict[Any, Any] = {}
DOCUMENT = Document(
    content="Airflow is a platform to programmatically author, schedule and monitor workflows."
)
ENCODING_TYPE = "UTF32"


class TestCloudNaturalLanguageHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.hook = CloudNaturalLanguageHook(gcp_conn_id="test")

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_credentials"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.natural_language.LanguageServiceClient")
    def test_language_service_client_creation(self, mock_client, mock_get_creds):
        result = self.hook.get_conn()
        mock_client.assert_called_once_with(credentials=mock_get_creds.return_value, client_info=CLIENT_INFO)
        assert mock_client.return_value == result
        assert self.hook._conn == result

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_analyze_entities(self, get_conn):
        get_conn.return_value.analyze_entities.return_value = API_RESPONSE
        result = self.hook.analyze_entities(document=DOCUMENT, encoding_type=ENCODING_TYPE)

        assert result == API_RESPONSE

        get_conn.return_value.analyze_entities.assert_called_once_with(
            document=DOCUMENT, encoding_type=ENCODING_TYPE, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_analyze_entity_sentiment(self, get_conn):
        get_conn.return_value.analyze_entity_sentiment.return_value = API_RESPONSE
        result = self.hook.analyze_entity_sentiment(document=DOCUMENT, encoding_type=ENCODING_TYPE)

        assert result == API_RESPONSE

        get_conn.return_value.analyze_entity_sentiment.assert_called_once_with(
            document=DOCUMENT, encoding_type=ENCODING_TYPE, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_analyze_sentiment(self, get_conn):
        get_conn.return_value.analyze_sentiment.return_value = API_RESPONSE
        result = self.hook.analyze_sentiment(document=DOCUMENT, encoding_type=ENCODING_TYPE)

        assert result == API_RESPONSE

        get_conn.return_value.analyze_sentiment.assert_called_once_with(
            document=DOCUMENT, encoding_type=ENCODING_TYPE, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_analyze_syntax(self, get_conn):
        get_conn.return_value.analyze_syntax.return_value = API_RESPONSE
        result = self.hook.analyze_syntax(document=DOCUMENT, encoding_type=ENCODING_TYPE)

        assert result == API_RESPONSE

        get_conn.return_value.analyze_syntax.assert_called_once_with(
            document=DOCUMENT, encoding_type=ENCODING_TYPE, retry=DEFAULT, timeout=None, metadata=()
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_annotate_text(self, get_conn):
        get_conn.return_value.annotate_text.return_value = API_RESPONSE
        result = self.hook.annotate_text(document=DOCUMENT, encoding_type=ENCODING_TYPE, features=None)

        assert result == API_RESPONSE

        get_conn.return_value.annotate_text.assert_called_once_with(
            document=DOCUMENT,
            encoding_type=ENCODING_TYPE,
            features=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.natural_language.CloudNaturalLanguageHook.get_conn",
    )
    def test_classify_text(self, get_conn):
        get_conn.return_value.classify_text.return_value = API_RESPONSE
        result = self.hook.classify_text(document=DOCUMENT)

        assert result == API_RESPONSE

        get_conn.return_value.classify_text.assert_called_once_with(
            document=DOCUMENT, retry=DEFAULT, timeout=None, metadata=()
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for testing Google Dataflow Beam Pipeline Operator with Python for Streaming job.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.apache.beam.hooks.beam import BeamRunnerType
from airflow.providers.apache.beam.operators.beam import BeamRunPythonPipelineOperator
from airflow.providers.google.cloud.operators.dataflow import DataflowStopJobOperator
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.operators.pubsub import (
    PubSubCreateTopicOperator,
    PubSubDeleteTopicOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "dataflow_native_python_streaming"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"

GCS_TMP = f"gs://{BUCKET_NAME}/temp/"
GCS_STAGING = f"gs://{BUCKET_NAME}/staging/"
GCS_PYTHON_SCRIPT = f"gs://{RESOURCE_DATA_BUCKET}/dataflow/python/streaming_wordcount.py"
LOCATION = "europe-west3"
TOPIC_ID = f"topic-{DAG_ID}"
TOPIC_ID_2 = f"topic-2-{DAG_ID}"

default_args = {
    "dataflow_default_options": {
        "tempLocation": GCS_TMP,
        "stagingLocation": GCS_STAGING,
    }
}

with DAG(
    DAG_ID,
    default_args=default_args,
    schedule="@once",
    start_date=datetime(2025, 1, 1),
    catchup=False,
    tags=["example", "dataflow"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    create_pub_sub_topic = PubSubCreateTopicOperator(
        task_id="create_topic", topic=TOPIC_ID, project_id=PROJECT_ID, fail_if_exists=False
    )
    create_pub_sub_topic_2 = PubSubCreateTopicOperator(
        task_id="create_topic_2", topic=TOPIC_ID_2, project_id=PROJECT_ID, fail_if_exists=False
    )

    # [START howto_operator_start_streaming_python_job]
    start_streaming_python_job = BeamRunPythonPipelineOperator(
        runner=BeamRunnerType.DataflowRunner,
        task_id="start_streaming_python_job",
        py_file=GCS_PYTHON_SCRIPT,
        py_options=[],
        pipeline_options={
            "temp_location": GCS_TMP,
            "input_topic": "projects/pubsub-public-data/topics/taxirides-realtime",
            "output_topic": f"projects/{PROJECT_ID}/topics/{TOPIC_ID}",
            "streaming": True,
        },
        py_requirements=["apache-beam[gcp]==2.67.0"],
        py_interpreter="python3",
        py_system_site_packages=False,
        dataflow_config={"location": LOCATION, "job_name": "start_python_job_streaming"},
    )
    # [END howto_operator_start_streaming_python_job]

    # [START howto_operator_start_streaming_python_job_deferrable]
    start_streaming_python_job_def = BeamRunPythonPipelineOperator(
        runner=BeamRunnerType.DataflowRunner,
        task_id="start_def_streaming_python_job",
        py_file=GCS_PYTHON_SCRIPT,
        py_options=[],
        pipeline_options={
            "temp_location": GCS_TMP,
            "input_topic": "projects/pubsub-public-data/topics/taxirides-realtime",
            "output_topic": f"projects/{PROJECT_ID}/topics/{TOPIC_ID_2}",
            "streaming": True,
        },
        py_requirements=["apache-beam[gcp]==2.67.0"],
        py_interpreter="python3",
        py_system_site_packages=False,
        dataflow_config={"location": LOCATION, "job_name": "start_python_job_streaming"},
        deferrable=True,
    )
    # [END howto_operator_start_streaming_python_job_deferrable]

    stop_dataflow_job = DataflowStopJobOperator(
        task_id="stop_dataflow_job",
        location=LOCATION,
        job_id="{{ task_instance.xcom_pull(task_ids='start_streaming_python_job')['dataflow_job_id'] }}",
    )

    stop_dataflow_job_deferrable = DataflowStopJobOperator(
        task_id="stop_dataflow_job_deferrable",
        location=LOCATION,
        job_id="{{ task_instance.xcom_pull(task_ids='start_def_streaming_python_job', key='dataflow_job_id') }}",
    )

    delete_topic = PubSubDeleteTopicOperator(task_id="delete_topic", topic=TOPIC_ID, project_id=PROJECT_ID)
    delete_topic.trigger_rule = TriggerRule.ALL_DONE

    delete_topic_2 = PubSubDeleteTopicOperator(
        task_id="delete_topic2", topic=TOPIC_ID_2, project_id=PROJECT_ID
    )
    delete_topic.trigger_rule = TriggerRule.ALL_DONE

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> create_pub_sub_topic
        >> create_pub_sub_topic_2
        # TEST BODY
        >> start_streaming_python_job
        >> start_streaming_python_job_def
        # TEST TEARDOWN
        >> [stop_dataflow_job, stop_dataflow_job_deferrable]
        >> delete_topic
        >> delete_topic_2
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from google.api_core.gapic_v1.client_info import ClientInfo

from airflow import version
from airflow.providers.google.common.consts import (
    CLIENT_INFO,
    GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
)


def test_google_default_deferrable_method_name():
    assert GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME == "execute_complete"


def test_client_info_instance():
    assert isinstance(CLIENT_INFO, ClientInfo)
    assert CLIENT_INFO.client_library_version == f"airflow_v{version.version}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example DAG using SQLToGoogleSheetsOperator.

This DAG relies on the following OS environment variables

* AIRFLOW__API__GOOGLE_KEY_PATH - Path to service account key file. Note, you can skip this variable if you
  run this DAG in a Composer environment.

"""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime
from typing import Any

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.common.sql.operators.sql import SQLExecuteQueryOperator
from airflow.providers.google.cloud.hooks.compute import ComputeEngineHook
from airflow.providers.google.cloud.hooks.compute_ssh import ComputeEngineSSHHook
from airflow.providers.google.cloud.operators.compute import (
    ComputeEngineDeleteInstanceOperator,
    ComputeEngineInsertInstanceOperator,
)
from airflow.providers.google.suite.operators.sheets import GoogleSheetsCreateSpreadsheetOperator
from airflow.providers.google.suite.transfers.sql_to_sheets import SQLToGoogleSheetsOperator
from airflow.providers.ssh.operators.ssh import SSHOperator
from airflow.providers.standard.operators.bash import BashOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google.gcp_api_client_helpers import create_airflow_connection, delete_airflow_connection

DAG_ID = "sql_to_sheets"
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "example-project")

IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))

REGION = "europe-west2"
ZONE = REGION + "-a"
NETWORK = "default"
CONNECTION_ID = f"pg_{DAG_ID}_{ENV_ID}".replace("-", "_")
CONNECTION_TYPE = "postgres"

BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
FILE_NAME = "result.json"

DB_NAME = "testdb"
DB_PORT = 5432
DB_USER_NAME = "root"
DB_USER_PASSWORD = "demo_password"
SETUP_POSTGRES_COMMAND = f"""
sudo apt update &&
sudo apt install -y docker.io &&
sudo docker run -d -p {DB_PORT}:{DB_PORT} --name {DB_NAME} \
    -e PGUSER={DB_USER_NAME} \
    -e POSTGRES_USER={DB_USER_NAME} \
    -e POSTGRES_PASSWORD={DB_USER_PASSWORD} \
    -e POSTGRES_DB={DB_NAME} \
    postgres
"""
SQL_TABLE = "test_table"
SQL_CREATE = f"CREATE TABLE IF NOT EXISTS {SQL_TABLE} (col_1 INT, col_2 VARCHAR(8))"
SQL_INSERT = f"INSERT INTO {SQL_TABLE} (col_1, col_2) VALUES (1, 'one'), (2, 'two')"
SQL_SELECT = f"SELECT * FROM {SQL_TABLE}"

GCE_MACHINE_TYPE = "n1-standard-1"
GCE_INSTANCE_NAME = f"instance-{DAG_ID}-{ENV_ID}".replace("_", "-")
GCE_INSTANCE_BODY = {
    "name": GCE_INSTANCE_NAME,
    "machine_type": f"zones/{ZONE}/machineTypes/{GCE_MACHINE_TYPE}",
    "disks": [
        {
            "boot": True,
            "auto_delete": True,
            "device_name": GCE_INSTANCE_NAME,
            "initialize_params": {
                "disk_size_gb": "10",
                "disk_type": f"zones/{ZONE}/diskTypes/pd-balanced",
                "source_image": "projects/debian-cloud/global/images/debian-12-bookworm-v20240611",
            },
        }
    ],
    "network_interfaces": [
        {
            "access_configs": [{"name": "External NAT", "network_tier": "PREMIUM"}],
            "stack_type": "IPV4_ONLY",
            "subnetwork": f"regions/{REGION}/subnetworks/default",
        }
    ],
}
FIREWALL_RULE_NAME = f"allow-http-{DB_PORT}-{DAG_ID}-{ENV_ID}".replace("_", "-")
CREATE_FIREWALL_RULE_COMMAND = f"""
if [ $AIRFLOW__API__GOOGLE_KEY_PATH ]; then \
 gcloud auth activate-service-account --key-file=$AIRFLOW__API__GOOGLE_KEY_PATH; \
fi;

if [ -z $(gcloud compute firewall-rules list --filter=name:{FIREWALL_RULE_NAME} --format="value(name)" --project={PROJECT_ID}) ]; then \
    gcloud compute firewall-rules create {FIREWALL_RULE_NAME} \
      --project={PROJECT_ID} \
      --direction=INGRESS \
      --priority=100 \
      --network={NETWORK} \
      --action=ALLOW \
      --rules=tcp:{DB_PORT} \
      --source-ranges=0.0.0.0/0
else
    echo "Firewall rule {FIREWALL_RULE_NAME} already exists."
fi
"""
DELETE_FIREWALL_RULE_COMMAND = f"""
if [ $AIRFLOW__API__GOOGLE_KEY_PATH ]; then \
 gcloud auth activate-service-account --key-file=$AIRFLOW__API__GOOGLE_KEY_PATH; \
fi; \
if [ $(gcloud compute firewall-rules list --filter=name:{FIREWALL_RULE_NAME} --format="value(name)" --project={PROJECT_ID}) ]; then \
    gcloud compute firewall-rules delete {FIREWALL_RULE_NAME} --project={PROJECT_ID} --quiet; \
fi;
"""

SHEETS_CONNECTION_ID = f"connection_{DAG_ID}_{ENV_ID}"
SPREADSHEET = {
    "properties": {"title": "Test1"},
    "sheets": [{"properties": {"title": "Sheet1"}}],
}

log = logging.getLogger(__name__)

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "postgres", "gcs"],
) as dag:
    create_gce_instance = ComputeEngineInsertInstanceOperator(
        task_id="create_gce_instance",
        project_id=PROJECT_ID,
        zone=ZONE,
        body=GCE_INSTANCE_BODY,
    )

    create_firewall_rule = BashOperator(
        task_id="create_firewall_rule",
        bash_command=CREATE_FIREWALL_RULE_COMMAND,
    )

    setup_postgres = SSHOperator(
        task_id="setup_postgres",
        ssh_hook=ComputeEngineSSHHook(
            user="username",
            instance_name=GCE_INSTANCE_NAME,
            zone=ZONE,
            project_id=PROJECT_ID,
            use_oslogin=False,
            use_iap_tunnel=False,
            cmd_timeout=180,
        ),
        command=SETUP_POSTGRES_COMMAND,
        retries=2,
    )

    @task
    def get_public_ip() -> str:
        hook = ComputeEngineHook()
        address = hook.get_instance_address(resource_id=GCE_INSTANCE_NAME, zone=ZONE, project_id=PROJECT_ID)
        return address

    get_public_ip_task = get_public_ip()

    @task
    def create_connection(connection_id: str, ip_address: str) -> None:
        connection: dict[str, Any] = {
            "description": "Example connection",
            "conn_type": CONNECTION_TYPE,
            "schema": DB_NAME,
            "host": ip_address,
            "login": DB_USER_NAME,
            "password": DB_USER_PASSWORD,
            "port": DB_PORT,
        }
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf=connection,
            is_composer=IS_COMPOSER,
        )

    create_connection_task = create_connection(connection_id=CONNECTION_ID, ip_address=get_public_ip_task)

    create_sql_table = SQLExecuteQueryOperator(
        task_id="create_sql_table",
        conn_id=CONNECTION_ID,
        sql=SQL_CREATE,
        retries=4,
    )

    insert_sql_data = SQLExecuteQueryOperator(
        task_id="insert_sql_data",
        conn_id=CONNECTION_ID,
        sql=SQL_INSERT,
    )

    @task
    def setup_sheets_connection():
        conn_extra = {
            "scope": "https://www.googleapis.com/auth/spreadsheets,https://www.googleapis.com/auth/cloud-platform",
            "project": PROJECT_ID,
            "keyfile_dict": "",  # Override to match your needs
        }
        conn_extra_json = json.dumps(conn_extra)
        connection: dict[str, Any] = {"conn_type": "google_cloud_platform", "extra": conn_extra_json}
        create_airflow_connection(
            connection_id=SHEETS_CONNECTION_ID,
            connection_conf=connection,
            is_composer=IS_COMPOSER,
        )

    setup_sheets_connection_task = setup_sheets_connection()

    create_spreadsheet = GoogleSheetsCreateSpreadsheetOperator(
        task_id="create_spreadsheet", spreadsheet=SPREADSHEET, gcp_conn_id=SHEETS_CONNECTION_ID
    )

    # [START upload_sql_to_sheets]
    upload_sql_to_sheet = SQLToGoogleSheetsOperator(
        task_id="upload_sql_to_sheet",
        sql=SQL_SELECT,
        sql_conn_id=CONNECTION_ID,
        database=DB_NAME,
        spreadsheet_id="{{ task_instance.xcom_pull(task_ids='create_spreadsheet', key='spreadsheet_id') }}",
        gcp_conn_id=SHEETS_CONNECTION_ID,
    )
    # [END upload_sql_to_sheets]

    delete_firewall_rule = BashOperator(
        task_id="delete_firewall_rule",
        bash_command=DELETE_FIREWALL_RULE_COMMAND,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_gce_instance = ComputeEngineDeleteInstanceOperator(
        task_id="delete_gce_instance",
        resource_id=GCE_INSTANCE_NAME,
        zone=ZONE,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    @task(task_id="delete_connection")
    def delete_connection(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id, is_composer=IS_COMPOSER)

    delete_connection_task = delete_connection(connection_id=CONNECTION_ID)
    delete_connection_sheets_task = delete_connection(connection_id=SHEETS_CONNECTION_ID)

    # TEST SETUP
    create_gce_instance >> get_public_ip_task >> create_connection_task
    [create_gce_instance, create_firewall_rule] >> setup_postgres
    [setup_postgres, create_connection_task, create_firewall_rule] >> create_sql_table >> insert_sql_data
    setup_sheets_connection_task >> create_spreadsheet

    (
        [create_spreadsheet, insert_sql_data]
        # TEST BODY
        >> upload_sql_to_sheet
    )

    # TEST TEARDOWN
    upload_sql_to_sheet >> [
        delete_firewall_rule,
        delete_gce_instance,
        delete_connection_task,
        delete_connection_sheets_task,
    ]

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use DisplayVideo.

In order to run this test, make sure you followed steps:
1. In your GCP project, create a service account that will be used to operate on Google Ads.
The name should be in format `display-video-service-account@{PROJECT_ID}.iam.gserviceaccount.com`
2. Generate a key for this service account and store it in the Secret Manager
under the name `google_display_video_service_account_key`.
3. Give this service account Editor permissions.
4. Make sure Google Display Video API is enabled in your GCP project.
5. Login to https://displayvideo.google.com/
6. In the Users section add your GCP service account to the list.
7. Store values of your advertiser id and GMP partner id to Secret Manager under names `google_display_video_advertiser_id`
and `google_display_video_gmp_partner_id`.
"""

from __future__ import annotations

import json
import os
from datetime import datetime

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.bigquery import BigQueryCreateEmptyDatasetOperator
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.providers.google.common.utils.get_secret import get_secret
from airflow.providers.google.marketing_platform.hooks.display_video import GoogleDisplayVideo360Hook
from airflow.providers.google.marketing_platform.operators.display_video import (
    GoogleDisplayVideo360CreateSDFDownloadTaskOperator,
    GoogleDisplayVideo360SDFtoGCSOperator,
)
from airflow.providers.google.marketing_platform.sensors.display_video import (
    GoogleDisplayVideo360GetSDFDownloadOperationSensor,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google.gcp_api_client_helpers import create_airflow_connection, delete_airflow_connection

DAG_ID = "display_video"
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
OBJECT_NAME = "files/report.csv"
PATH_TO_UPLOAD_FILE = os.environ.get("GCP_GCS_PATH_TO_UPLOAD_FILE", "test-gcs-example.csv")
BUCKET_FILE_LOCATION = PATH_TO_UPLOAD_FILE.rpartition("/")[-1]

IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))

CONNECTION_TYPE = "google_cloud_platform"
CONN_ID = "google_display_video_default"

DISPLAY_VIDEO_ADVERTISER_ID = "google_display_video_advertiser_id"
DISPLAY_VIDEO_GMP_PARTNER_ID = "google_display_video_gmp_partner_id"
DISPLAY_VIDEO_SERVICE_ACCOUNT_KEY = "google_display_video_service_account_key"

# [START howto_display_video_env_variables]
ADVERTISER_ID = "{{ task_instance.xcom_pull('get_display_video_advertiser_id') }}"
GMP_PARTNER_ID = "{{ task_instance.xcom_pull('get_display_video_gmp_partner_id') }}"
SDF_VERSION = "SDF_VERSION_8_1"
BQ_DATASET = f"bq_dataset_{DAG_ID}_{ENV_ID}".replace("-", "_")
ENTITY_TYPE = "CAMPAIGN"
ERF_SOURCE_OBJECT = GoogleDisplayVideo360Hook.erf_uri(GMP_PARTNER_ID, ENTITY_TYPE)

CREATE_SDF_DOWNLOAD_TASK_BODY_REQUEST: dict = {
    "version": SDF_VERSION,
    "advertiserId": ADVERTISER_ID,
    "inventorySourceFilter": {"inventorySourceIds": []},
}

# [END howto_display_video_env_variables]


with DAG(
    "display_video_sdf",
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "display_video_sdf"],
    render_template_as_native_obj=True,
) as dag:

    @task
    def get_display_video_advertiser_id():
        return get_secret(secret_id=DISPLAY_VIDEO_ADVERTISER_ID).strip()

    get_display_video_advertiser_id_task = get_display_video_advertiser_id()

    @task
    def get_display_video_gmp_partner_id():
        return get_secret(secret_id=DISPLAY_VIDEO_GMP_PARTNER_ID).strip()

    get_display_video_gmp_partner_id_task = get_display_video_gmp_partner_id()

    @task
    def get_display_video_service_account_key():
        return get_secret(secret_id=DISPLAY_VIDEO_SERVICE_ACCOUNT_KEY)

    get_display_video_service_account_key_task = get_display_video_service_account_key()

    @task
    def create_connection_display_video(connection_id: str, key) -> None:
        conn_extra_json = json.dumps(
            {
                "keyfile_dict": key,
                "project": PROJECT_ID,
                "scope": "https://www.googleapis.com/auth/display-video, https://www.googleapis.com/auth/cloud-platform",
            }
        )
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf={"conn_type": CONNECTION_TYPE, "extra": conn_extra_json},
            is_composer=IS_COMPOSER,
        )

    create_connection_display_video_task = create_connection_display_video(
        connection_id=CONN_ID, key=get_display_video_service_account_key_task
    )

    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket", bucket_name=BUCKET_NAME, project_id=PROJECT_ID, gcp_conn_id=CONN_ID
    )

    # [START howto_google_display_video_create_sdf_download_task_operator]
    create_sdf_download_task = GoogleDisplayVideo360CreateSDFDownloadTaskOperator(
        task_id="create_sdf_download_task",
        body_request=CREATE_SDF_DOWNLOAD_TASK_BODY_REQUEST,
        gcp_conn_id=CONN_ID,
    )
    operation_name = '{{ task_instance.xcom_pull("create_sdf_download_task")["name"] }}'
    # [END howto_google_display_video_create_sdf_download_task_operator]

    # [START howto_google_display_video_wait_for_operation_sensor]
    wait_for_operation = GoogleDisplayVideo360GetSDFDownloadOperationSensor(
        task_id="wait_for_operation", operation_name=operation_name, gcp_conn_id=CONN_ID
    )
    # [END howto_google_display_video_wait_for_operation_sensor]

    # [START howto_google_display_video_save_sdf_in_gcs_operator]
    save_sdf_in_gcs = GoogleDisplayVideo360SDFtoGCSOperator(
        task_id="save_sdf_in_gcs",
        operation_name=operation_name,
        bucket_name=BUCKET_NAME,
        object_name=BUCKET_FILE_LOCATION,
        gzip=False,
        gcp_conn_id=CONN_ID,
    )
    # [END howto_google_display_video_save_sdf_in_gcs_operator]

    create_dataset = BigQueryCreateEmptyDatasetOperator(
        task_id="create_dataset", dataset_id=BQ_DATASET, gcp_conn_id=CONN_ID
    )

    # [START howto_google_display_video_gcs_to_big_query_operator]
    upload_sdf_to_big_query = GCSToBigQueryOperator(
        task_id="upload_sdf_to_big_query",
        bucket=BUCKET_NAME,
        source_objects=[PATH_TO_UPLOAD_FILE],
        destination_project_dataset_table=f"{BQ_DATASET}.gcs_to_bq_table",
        schema_fields=[
            {"name": "name", "type": "STRING", "mode": "NULLABLE"},
            {"name": "post_abbr", "type": "STRING", "mode": "NULLABLE"},
        ],
        write_disposition="WRITE_TRUNCATE",
        gcp_conn_id=CONN_ID,
    )
    # [END howto_google_display_video_gcs_to_big_query_operator]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=BUCKET_NAME,
        gcp_conn_id=CONN_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    @task(task_id="delete_connection_task")
    def delete_connection_display_video(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id, is_composer=IS_COMPOSER)

    delete_connection_task = delete_connection_display_video(connection_id=CONN_ID)

    (
        # TEST SETUP
        [
            get_display_video_advertiser_id_task,
            get_display_video_gmp_partner_id_task,
            get_display_video_service_account_key_task,
        ]
        >> create_connection_display_video_task
        >> create_bucket
        >> create_dataset
        # TEST BODY
        >> create_sdf_download_task
        >> wait_for_operation
        >> save_sdf_in_gcs
        >> upload_sdf_to_big_query
        # TEST TEARDOWN
        >> delete_bucket
        >> delete_connection_task
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import Mock

import pytest

ray = pytest.importorskip("ray")

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.sensors.vertex_ai.feature_store import FeatureViewSyncSensor

TASK_ID = "test-task"
GCP_CONN_ID = "test-conn"
GCP_LOCATION = "us-central1"
FEATURE_VIEW_SYNC_NAME = "projects/123/locations/us-central1/featureViews/test-view/operations/sync-123"
TIMEOUT = 120


class TestFeatureViewSyncSensor:
    def create_sync_response(self, end_time=None, row_synced=None, total_slot=None):
        response = {}
        if end_time is not None:
            response["end_time"] = end_time
        if row_synced is not None and total_slot is not None:
            response["sync_summary"] = {"row_synced": str(row_synced), "total_slot": str(total_slot)}
        return response

    @mock.patch("airflow.providers.google.cloud.sensors.vertex_ai.feature_store.FeatureStoreHook")
    def test_sync_completed(self, mock_hook):
        mock_hook.return_value.get_feature_view_sync.return_value = self.create_sync_response(
            end_time=1234567890, row_synced=1000, total_slot=5
        )

        sensor = FeatureViewSyncSensor(
            task_id=TASK_ID,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
            location=GCP_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            timeout=TIMEOUT,
        )
        ret = sensor.poke(context={})

        mock_hook.return_value.get_feature_view_sync.assert_called_once_with(
            location=GCP_LOCATION,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
        )
        assert ret

    @mock.patch("airflow.providers.google.cloud.sensors.vertex_ai.feature_store.FeatureStoreHook")
    def test_sync_running(self, mock_hook):
        mock_hook.return_value.get_feature_view_sync.return_value = self.create_sync_response(
            end_time=0, row_synced=0, total_slot=5
        )

        sensor = FeatureViewSyncSensor(
            task_id=TASK_ID,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
            location=GCP_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            timeout=TIMEOUT,
        )
        ret = sensor.poke(context={})

        mock_hook.return_value.get_feature_view_sync.assert_called_once_with(
            location=GCP_LOCATION,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
        )
        assert not ret

    @mock.patch("airflow.providers.google.cloud.sensors.vertex_ai.feature_store.FeatureStoreHook")
    def test_sync_error_with_retry(self, mock_hook):
        mock_hook.return_value.get_feature_view_sync.side_effect = Exception("API Error")

        sensor = FeatureViewSyncSensor(
            task_id=TASK_ID,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
            location=GCP_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            timeout=TIMEOUT,
        )
        ret = sensor.poke(context={})

        mock_hook.return_value.get_feature_view_sync.assert_called_once_with(
            location=GCP_LOCATION,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
        )
        assert not ret

    @mock.patch("airflow.providers.google.cloud.sensors.vertex_ai.feature_store.FeatureStoreHook")
    def test_timeout_during_running(self, mock_hook):
        mock_hook.return_value.get_feature_view_sync.return_value = self.create_sync_response(
            end_time=0, row_synced=0, total_slot=5
        )

        sensor = FeatureViewSyncSensor(
            task_id=TASK_ID,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
            location=GCP_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            timeout=TIMEOUT,
            wait_timeout=300,
        )

        sensor._duration = Mock()
        sensor._duration.return_value = 301

        with pytest.raises(
            AirflowException,
            match=f"Timeout: Feature View sync {FEATURE_VIEW_SYNC_NAME} not completed after 300s",
        ):
            sensor.poke(context={})

    @mock.patch("airflow.providers.google.cloud.sensors.vertex_ai.feature_store.FeatureStoreHook")
    def test_timeout_during_error(self, mock_hook):
        mock_hook.return_value.get_feature_view_sync.side_effect = Exception("API Error")

        sensor = FeatureViewSyncSensor(
            task_id=TASK_ID,
            feature_view_sync_name=FEATURE_VIEW_SYNC_NAME,
            location=GCP_LOCATION,
            gcp_conn_id=GCP_CONN_ID,
            timeout=TIMEOUT,
            wait_timeout=300,
        )

        sensor._duration = Mock()
        sensor._duration.return_value = 301

        with pytest.raises(
            AirflowException,
            match=f"Timeout: Feature View sync {FEATURE_VIEW_SYNC_NAME} not completed after 300s",
        ):
            sensor.poke(context={})



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use Google Analytics (GA4) Admin Operators.

This DAG relies on the following OS environment variables

* GA_ACCOUNT_ID - Google Analytics account id.
* GA_GOOGLE_ADS_PROPERTY_ID - Google Analytics property's id associated with Google Ads Link.

In order to run this test, make sure you followed steps:
1. Login to https://analytics.google.com
2. In the settings section create an account and save its ID in the Google Cloud Secret Manager with id
saved in the constant GOOGLE_ANALYTICS_ACCOUNT_SECRET_ID.
3. In the settings section go to the Property access management page and add your service account email with
Editor permissions. This service account should be created on behalf of the account from the step 1.
4. Make sure Google Analytics Admin API is enabled in your GCP project.
5. Create Google Ads account and link it to your Google Analytics account in the GA admin panel.
6. Associate the Google Ads account with a property, and save this property's id in the Google Cloud Secret
Manager with id saved in the constant GOOGLE_ADS_PROPERTY_ID.
"""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime
from typing import Any

from google.analytics import admin_v1beta as google_analytics

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.common.utils.get_secret import get_secret
from airflow.providers.google.marketing_platform.operators.analytics_admin import (
    GoogleAnalyticsAdminCreateDataStreamOperator,
    GoogleAnalyticsAdminCreatePropertyOperator,
    GoogleAnalyticsAdminDeleteDataStreamOperator,
    GoogleAnalyticsAdminDeletePropertyOperator,
    GoogleAnalyticsAdminGetGoogleAdsLinkOperator,
    GoogleAnalyticsAdminListAccountsOperator,
    GoogleAnalyticsAdminListGoogleAdsLinksOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google.gcp_api_client_helpers import create_airflow_connection, delete_airflow_connection

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "google_analytics_admin"
IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))

CONNECTION_ID = f"connection_{DAG_ID}_{ENV_ID}"
GOOGLE_ANALYTICS_ACCOUNT_SECRET_ID = "google_analytics_account_id"
GOOGLE_ADS_PROPERTY_SECRET_ID = "google_ads_property_id"
PROPERTY_ID = "{{ task_instance.xcom_pull('create_property')['name'].split('/')[-1] }}"
DATA_STREAM_ID = "{{ task_instance.xcom_pull('create_data_stream')['name'].split('/')[-1] }}"
GA_ADS_LINK_ID = "{{ task_instance.xcom_pull('list_google_ads_links')[0]['name'].split('/')[-1] }}"

log = logging.getLogger(__name__)


with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs,
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "analytics"],
) as dag:

    @task
    def create_connection(connection_id: str) -> None:
        conn_extra_json = json.dumps(
            {
                "scope": "https://www.googleapis.com/auth/analytics.edit,"
                "https://www.googleapis.com/auth/analytics.readonly",
            }
        )
        connection: dict[str, Any] = {"conn_type": "google_cloud_platform", "extra": conn_extra_json}
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf=connection,
            is_composer=IS_COMPOSER,
        )

    create_connection_task = create_connection(connection_id=CONNECTION_ID)

    @task
    def get_google_analytics_account_id():
        return get_secret(secret_id=GOOGLE_ANALYTICS_ACCOUNT_SECRET_ID)

    get_google_analytics_account_id_task = get_google_analytics_account_id()

    @task
    def get_google_ads_property_id():
        return get_secret(secret_id=GOOGLE_ADS_PROPERTY_SECRET_ID)

    get_google_ads_property_id_task = get_google_ads_property_id()

    # [START howto_marketing_platform_list_accounts_operator]
    list_accounts = GoogleAnalyticsAdminListAccountsOperator(
        task_id="list_account",
        gcp_conn_id=CONNECTION_ID,
        show_deleted=True,
    )
    # [END howto_marketing_platform_list_accounts_operator]

    # [START howto_marketing_platform_create_property_operator]
    create_property = GoogleAnalyticsAdminCreatePropertyOperator(
        task_id="create_property",
        analytics_property={
            "parent": f"accounts/{get_google_analytics_account_id_task}",
            "display_name": "Test display name",
            "time_zone": "America/Los_Angeles",
        },
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_create_property_operator]

    # [START howto_marketing_platform_create_data_stream_operator]
    create_data_stream = GoogleAnalyticsAdminCreateDataStreamOperator(
        task_id="create_data_stream",
        property_id=PROPERTY_ID,
        data_stream={
            "display_name": "Test data stream",
            "web_stream_data": {
                "default_uri": "www.example.com",
            },
            "type_": google_analytics.DataStream.DataStreamType.WEB_DATA_STREAM,
        },
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_create_data_stream_operator]

    # [START howto_marketing_platform_delete_data_stream_operator]
    delete_data_stream = GoogleAnalyticsAdminDeleteDataStreamOperator(
        task_id="delete_datastream",
        property_id=PROPERTY_ID,
        data_stream_id=DATA_STREAM_ID,
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_delete_data_stream_operator]

    # [START howto_marketing_platform_delete_property_operator]
    delete_property = GoogleAnalyticsAdminDeletePropertyOperator(
        task_id="delete_property",
        property_id=PROPERTY_ID,
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_delete_property_operator]
    delete_property.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_marketing_platform_list_google_ads_links]
    list_google_ads_links = GoogleAnalyticsAdminListGoogleAdsLinksOperator(
        task_id="list_google_ads_links",
        property_id=get_google_ads_property_id_task,
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_list_google_ads_links]

    # [START howto_marketing_platform_get_google_ad_link]
    get_ad_link = GoogleAnalyticsAdminGetGoogleAdsLinkOperator(
        task_id="get_ad_link",
        property_id=get_google_ads_property_id_task,
        google_ads_link_id=GA_ADS_LINK_ID,
        gcp_conn_id=CONNECTION_ID,
    )
    # [END howto_marketing_platform_get_google_ad_link]

    @task(task_id="delete_connection")
    def delete_connection(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id, is_composer=IS_COMPOSER)

    delete_connection_task = delete_connection(connection_id=CONNECTION_ID)

    (
        # TEST SETUP
        [create_connection_task, get_google_analytics_account_id_task, get_google_ads_property_id_task]
        # TEST BODY
        >> list_accounts
        >> create_property
        >> create_data_stream
        >> delete_data_stream
        >> delete_property
        >> list_google_ads_links
        >> get_ad_link
        # TEST TEARDOWN
        >> delete_connection_task
    )
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google CampaignManager operators."""

from __future__ import annotations

import json
import tempfile
import uuid
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from googleapiclient import http

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.marketing_platform.hooks.campaign_manager import GoogleCampaignManagerHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class GoogleCampaignManagerDeleteReportOperator(BaseOperator):
    """
    Deletes a report by its ID.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/reports/delete`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerDeleteReportOperator`

    :param profile_id: The DFA user profile ID.
    :param report_name: The name of the report to delete.
    :param report_id: The ID of the report.
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "report_id",
        "report_name",
        "api_version",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        profile_id: str,
        report_name: str | None = None,
        report_id: str | None = None,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if not (report_name or report_id):
            raise AirflowException("Please provide `report_name` or `report_id`.")
        if report_name and report_id:
            raise AirflowException("Please provide only one parameter `report_name` or `report_id`.")

        self.profile_id = profile_id
        self.report_name = report_name
        self.report_id = report_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        if self.report_name:
            reports = hook.list_reports(profile_id=self.profile_id)
            reports_with_name = [r for r in reports if r["name"] == self.report_name]
            for report in reports_with_name:
                report_id = report["id"]
                self.log.info("Deleting Campaign Manager report: %s", report_id)
                hook.delete_report(profile_id=self.profile_id, report_id=report_id)
                self.log.info("Report deleted.")
        elif self.report_id:
            self.log.info("Deleting Campaign Manager report: %s", self.report_id)
            hook.delete_report(profile_id=self.profile_id, report_id=self.report_id)
            self.log.info("Report deleted.")


class GoogleCampaignManagerDownloadReportOperator(BaseOperator):
    """
    Retrieves a report and uploads it to GCS bucket.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/reports/get`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerDownloadReportOperator`

    :param profile_id: The DFA user profile ID.
    :param report_id: The ID of the report.
    :param file_id: The ID of the report file.
    :param bucket_name: The bucket to upload to.
    :param report_name: The report name to set when uploading the local file.
    :param gzip: Option to compress local file or file data for upload
    :param chunk_size: File will be downloaded in chunks of this many bytes.
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "report_id",
        "file_id",
        "bucket_name",
        "report_name",
        "chunk_size",
        "api_version",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        profile_id: str,
        report_id: str,
        file_id: str,
        bucket_name: str,
        report_name: str | None = None,
        gzip: bool = True,
        chunk_size: int = 10 * 1024 * 1024,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.profile_id = profile_id
        self.report_id = report_id
        self.file_id = file_id
        self.api_version = api_version
        self.chunk_size = chunk_size
        self.gzip = gzip
        self.bucket_name = bucket_name
        self.report_name = report_name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def _resolve_file_name(self, name: str) -> str:
        csv = ".csv"
        gzip = ".gz"
        if not name.endswith(csv):
            name += csv
        if self.gzip:
            name += gzip
        return name

    @staticmethod
    def _set_bucket_name(name: str) -> str:
        bucket = name if not name.startswith("gs://") else name[5:]
        return bucket.strip("/")

    def execute(self, context: Context) -> None:
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        # Get name of the report
        report = hook.get_report(file_id=self.file_id, profile_id=self.profile_id, report_id=self.report_id)
        report_name = self.report_name or report.get("fileName", str(uuid.uuid4()))
        report_name = self._resolve_file_name(report_name)

        # Download the report
        self.log.info("Starting downloading report %s", self.report_id)
        request = hook.get_report_file(
            profile_id=self.profile_id, report_id=self.report_id, file_id=self.file_id
        )
        with tempfile.NamedTemporaryFile() as temp_file:
            downloader = http.MediaIoBaseDownload(fd=temp_file, request=request, chunksize=self.chunk_size)
            download_finished = False
            while not download_finished:
                _, download_finished = downloader.next_chunk()

            temp_file.flush()
            # Upload the local file to bucket
            bucket_name = self._set_bucket_name(self.bucket_name)
            gcs_hook.upload(
                bucket_name=bucket_name,
                object_name=report_name,
                gzip=self.gzip,
                filename=temp_file.name,
                mime_type="text/csv",
            )

        context["task_instance"].xcom_push(key="report_name", value=report_name)


class GoogleCampaignManagerInsertReportOperator(BaseOperator):
    """
    Creates a report.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/reports/insert`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerInsertReportOperator`

    :param profile_id: The DFA user profile ID.
    :param report: Report to be created.
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "report",
        "api_version",
        "gcp_conn_id",
        "impersonation_chain",
    )

    template_ext: Sequence[str] = (".json",)

    def __init__(
        self,
        *,
        profile_id: str,
        report: dict[str, Any],
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.profile_id = profile_id
        self.report = report
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def prepare_template(self) -> None:
        # If .json is passed then we have to read the file
        if isinstance(self.report, str) and self.report.endswith(".json"):
            with open(self.report) as file:
                self.report = json.load(file)

    def execute(self, context: Context):
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Inserting Campaign Manager report.")
        response = hook.insert_report(profile_id=self.profile_id, report=self.report)
        report_id = response.get("id")
        context["task_instance"].xcom_push(key="report_id", value=report_id)
        self.log.info("Report successfully inserted. Report id: %s", report_id)
        return response


class GoogleCampaignManagerRunReportOperator(BaseOperator):
    """
    Runs a report.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/reports/run`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerRunReportOperator`

    :param profile_id: The DFA profile ID.
    :param report_id: The ID of the report.
    :param synchronous: If set and true, tries to run the report synchronously.
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "report_id",
        "synchronous",
        "api_version",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        profile_id: str,
        report_id: str,
        synchronous: bool = False,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.profile_id = profile_id
        self.report_id = report_id
        self.synchronous = synchronous
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Running report %s", self.report_id)
        response = hook.run_report(
            profile_id=self.profile_id,
            report_id=self.report_id,
            synchronous=self.synchronous,
        )
        file_id = response.get("id")
        context["task_instance"].xcom_push(key="file_id", value=file_id)
        self.log.info("Report file id: %s", file_id)
        return response


class GoogleCampaignManagerBatchInsertConversionsOperator(BaseOperator):
    """
    Inserts conversions.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/conversions/batchinsert`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerBatchInsertConversionsOperator`

    :param profile_id: User profile ID associated with this request.
    :param conversions: Conversions to insert, should be type of Conversion:
        https://developers.google.com/doubleclick-advertisers/rest/v4/conversions
    :param encryption_entity_type: The encryption entity type. This should match the encryption
        configuration for ad serving or Data Transfer.
    :param encryption_entity_id: The encryption entity ID. This should match the encryption
        configuration for ad serving or Data Transfer.
    :param encryption_source: Describes whether the encrypted cookie was received from ad serving
        (the %m macro) or from Data Transfer.
    :param max_failed_inserts: The maximum number of conversions that failed to be inserted
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "conversions",
        "encryption_entity_type",
        "encryption_entity_id",
        "encryption_source",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        profile_id: str,
        conversions: list[dict[str, Any]],
        encryption_entity_type: str,
        encryption_entity_id: int,
        encryption_source: str,
        max_failed_inserts: int = 0,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.profile_id = profile_id
        self.conversions = conversions
        self.encryption_entity_type = encryption_entity_type
        self.encryption_entity_id = encryption_entity_id
        self.encryption_source = encryption_source
        self.max_failed_inserts = max_failed_inserts
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        response = hook.conversions_batch_insert(
            profile_id=self.profile_id,
            conversions=self.conversions,
            encryption_entity_type=self.encryption_entity_type,
            encryption_entity_id=self.encryption_entity_id,
            encryption_source=self.encryption_source,
            max_failed_inserts=self.max_failed_inserts,
        )
        return response


class GoogleCampaignManagerBatchUpdateConversionsOperator(BaseOperator):
    """
    Updates existing conversions.

    .. seealso::
        Check official API docs:
        `https://developers.google.com/doubleclick-advertisers/rest/v4/conversions/batchupdate`

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleCampaignManagerBatchUpdateConversionsOperator`

    :param profile_id: User profile ID associated with this request.
    :param conversions: Conversations to update, should be type of Conversion:
        https://developers.google.com/doubleclick-advertisers/rest/v4/conversions
    :param encryption_entity_type: The encryption entity type. This should match the encryption
        configuration for ad serving or Data Transfer.
    :param encryption_entity_id: The encryption entity ID. This should match the encryption
        configuration for ad serving or Data Transfer.
    :param encryption_source: Describes whether the encrypted cookie was received from ad serving
        (the %m macro) or from Data Transfer.
    :param max_failed_updates: The maximum number of conversions that failed to be updated
    :param api_version: The version of the api that will be requested, for example 'v4'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "profile_id",
        "conversions",
        "encryption_entity_type",
        "encryption_entity_id",
        "encryption_source",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        profile_id: str,
        conversions: list[dict[str, Any]],
        encryption_entity_type: str,
        encryption_entity_id: int,
        encryption_source: str,
        max_failed_updates: int = 0,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.profile_id = profile_id
        self.conversions = conversions
        self.encryption_entity_type = encryption_entity_type
        self.encryption_entity_id = encryption_entity_id
        self.encryption_source = encryption_source
        self.max_failed_updates = max_failed_updates
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook = GoogleCampaignManagerHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        response = hook.conversions_batch_update(
            profile_id=self.profile_id,
            conversions=self.conversions,
            encryption_entity_type=self.encryption_entity_type,
            encryption_entity_id=self.encryption_entity_id,
            encryption_source=self.encryption_source,
            max_failed_updates=self.max_failed_updates,
        )
        return response



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Managed Service for Apache Kafka hook."""

from __future__ import annotations

import base64
import datetime
import json
import time
from collections.abc import Sequence
from copy import deepcopy
from typing import TYPE_CHECKING

from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.auth.transport import requests as google_requests
from google.cloud.managedkafka_v1 import Cluster, ConsumerGroup, ManagedKafkaClient, Topic, types

from airflow.exceptions import AirflowException
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

if TYPE_CHECKING:
    from google.api_core.operation import Operation
    from google.api_core.retry import Retry
    from google.auth.credentials import Credentials
    from google.cloud.managedkafka_v1.services.managed_kafka.pagers import (
        ListClustersPager,
        ListConsumerGroupsPager,
        ListTopicsPager,
    )
    from google.protobuf.field_mask_pb2 import FieldMask


class ManagedKafkaTokenProvider:
    """Helper for providing authentication token for establishing connection via confluent to Apache Kafka cluster managed by Google Cloud."""

    def __init__(
        self,
        credentials: Credentials,
    ):
        self._credentials = credentials
        self._header = json.dumps(dict(typ="JWT", alg="GOOG_OAUTH2_TOKEN"))

    def _valid_credentials(self):
        if not self._credentials.valid:
            self._credentials.refresh(google_requests.Request())
            return self._credentials

    def _get_jwt(self, credentials):
        return json.dumps(
            dict(
                exp=credentials.expiry.timestamp(),
                iss="Google",
                iat=datetime.datetime.now(datetime.timezone.utc).timestamp(),
                scope="kafka",
                sub=credentials.service_account_email,
            )
        )

    def _b64_encode(self, source):
        return base64.urlsafe_b64encode(source.encode("utf-8")).decode("utf-8").rstrip("=")

    def _get_kafka_access_token(self, credentials):
        return ".".join(
            [
                self._b64_encode(self._header),
                self._b64_encode(self._get_jwt(credentials)),
                self._b64_encode(credentials.token),
            ]
        )

    def confluent_token(self):
        credentials = self._valid_credentials()

        utc_expiry = credentials.expiry.replace(tzinfo=datetime.timezone.utc)
        expiry_seconds = (utc_expiry - datetime.datetime.now(datetime.timezone.utc)).total_seconds()

        return self._get_kafka_access_token(credentials), time.time() + expiry_seconds


class ManagedKafkaHook(GoogleBaseHook):
    """Hook for Managed Service for Apache Kafka APIs."""

    def __init__(
        self,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(gcp_conn_id, impersonation_chain, **kwargs)

    def get_managed_kafka_client(self) -> ManagedKafkaClient:
        """Return ManagedKafkaClient object."""
        return ManagedKafkaClient(
            credentials=self.get_credentials(),
            client_info=CLIENT_INFO,
        )

    def wait_for_operation(self, operation: Operation, timeout: float | None = None):
        """Wait for long-lasting operation to complete."""
        try:
            return operation.result(timeout=timeout)
        except Exception:
            error = operation.exception(timeout=timeout)
            raise AirflowException(error)

    def get_confluent_token(self, config_str: str):
        """Get the authentication token for confluent client."""
        token_provider = ManagedKafkaTokenProvider(credentials=self.get_credentials())
        token = token_provider.confluent_token()
        return token

    @GoogleBaseHook.fallback_to_default_project_id
    def create_cluster(
        self,
        project_id: str,
        location: str,
        cluster: types.Cluster | dict,
        cluster_id: str,
        request_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create a new Apache Kafka cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster: Required. Configuration of the cluster to create. Its ``name`` field is ignored.
        :param cluster_id: Required. The ID to use for the cluster, which will become the final component of
            the cluster's name. The ID must be 1-63 characters long, and match the regular expression
            ``[a-z]([-a-z0-9]*[a-z0-9])?`` to comply with RFC 1035. This value is structured like: ``my-cluster-id``.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            to avoid duplication of requests. If a request times out or fails, retrying with the same ID
            allows the server to recognize the previous attempt. For at least 60 minutes, the server ignores
            duplicate requests bearing the same ID. For example, consider a situation where you make an
            initial request and the request times out. If you make the request again with the same request ID
            within 60 minutes of the last request, the server checks if an original operation with the same
            request ID was received. If so, the server ignores the second request. The request ID must be a
            valid UUID. A zero UUID is not supported (00000000-0000-0000-0000-000000000000).
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        parent = client.common_location_path(project_id, location)

        operation = client.create_cluster(
            request={
                "parent": parent,
                "cluster_id": cluster_id,
                "cluster": cluster,
                "request_id": request_id,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return operation

    @GoogleBaseHook.fallback_to_default_project_id
    def list_clusters(
        self,
        project_id: str,
        location: str,
        page_size: int | None = None,
        page_token: str | None = None,
        filter: str | None = None,
        order_by: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListClustersPager:
        """
        List the clusters in a given project and location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param page_size: Optional. The maximum number of clusters to return. The service may return fewer
            than this value. If unspecified, server will pick an appropriate default.
        :param page_token: Optional. A page token, received from a previous ``ListClusters`` call. Provide
            this to retrieve the subsequent page.
            When paginating, all other parameters provided to ``ListClusters`` must match the call that
            provided the page token.
        :param filter: Optional. Filter expression for the result.
        :param order_by: Optional. Order by fields for the result.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        parent = client.common_location_path(project_id, location)

        result = client.list_clusters(
            request={
                "parent": parent,
                "page_size": page_size,
                "page_token": page_token,
                "filter": filter,
                "order_by": order_by,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.Cluster:
        """
        Return the properties of a single cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose configuration to return.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.cluster_path(project_id, location, cluster_id)

        result = client.get_cluster(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        cluster: types.Cluster | dict,
        update_mask: FieldMask | dict,
        request_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Update the properties of a single cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose configuration to update.
        :param cluster: Required. The cluster to update.
        :param update_mask: Required. Field mask is used to specify the fields to be overwritten in the
            cluster resource by the update. The fields specified in the update_mask are relative to the
            resource, not the full request. A field will be overwritten if it is in the mask.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            to avoid duplication of requests. If a request times out or fails, retrying with the same ID
            allows the server to recognize the previous attempt. For at least 60 minutes, the server ignores
            duplicate requests bearing the same ID.
            For example, consider a situation where you make an initial request and the request times out. If
            you make the request again with the same request ID within 60 minutes of the last request, the
            server checks if an original operation with the same request ID was received. If so, the server
            ignores the second request.
            The request ID must be a valid UUID. A zero UUID is not supported (00000000-0000-0000-0000-000000000000).
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        _cluster = deepcopy(cluster) if isinstance(cluster, dict) else Cluster.to_dict(cluster)
        _cluster["name"] = client.cluster_path(project_id, location, cluster_id)

        operation = client.update_cluster(
            request={
                "update_mask": update_mask,
                "cluster": _cluster,
                "request_id": request_id,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return operation

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_cluster(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        request_id: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete a single cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster to delete.
        :param request_id: Optional. An optional request ID to identify requests. Specify a unique request ID
            to avoid duplication of requests. If a request times out or fails, retrying with the same ID
            allows the server to recognize the previous attempt. For at least 60 minutes, the server ignores
            duplicate requests bearing the same ID.
            For example, consider a situation where you make an initial request and the request times out. If
            you make the request again with the same request ID within 60 minutes of the last request, the
            server checks if an original operation with the same request ID was received. If so, the server
            ignores the second request.
            The request ID must be a valid UUID. A zero UUID is not supported (00000000-0000-0000-0000-000000000000).
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.cluster_path(project_id, location, cluster_id)

        operation = client.delete_cluster(
            request={
                "name": name,
                "request_id": request_id,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return operation

    @GoogleBaseHook.fallback_to_default_project_id
    def create_topic(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        topic_id: str,
        topic: types.Topic | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.Topic:
        """
        Create a new topic in a given project and location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster in which to create the topic.
        :param topic_id: Required. The ID to use for the topic, which will become the final component of the
            topic's name.
        :param topic: Required. Configuration of the topic to create.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        parent = client.cluster_path(project_id, location, cluster_id)

        result = client.create_topic(
            request={
                "parent": parent,
                "topic_id": topic_id,
                "topic": topic,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_topics(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        page_size: int | None = None,
        page_token: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListTopicsPager:
        """
        List the topics in a given cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose topics are to be listed.
        :param page_size: Optional. The maximum number of topics to return. The service may return fewer than
            this value. If unset or zero, all topics for the parent is returned.
        :param page_token: Optional. A page token, received from a previous ``ListTopics`` call. Provide this
            to retrieve the subsequent page. When paginating, all other parameters provided to ``ListTopics``
            must match the call that provided the page token.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        parent = client.cluster_path(project_id, location, cluster_id)

        result = client.list_topics(
            request={
                "parent": parent,
                "page_size": page_size,
                "page_token": page_token,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_topic(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        topic_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.Topic:
        """
        Return the properties of a single topic.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose topic is to be returned.
        :param topic_id: Required. The ID of the topic whose configuration to return.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.topic_path(project_id, location, cluster_id, topic_id)

        result = client.get_topic(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_topic(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        topic_id: str,
        topic: types.Topic | dict,
        update_mask: FieldMask | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.Topic:
        """
        Update the properties of a single topic.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose topic is to be updated.
        :param topic_id: Required. The ID of the topic whose configuration to update.
        :param topic: Required. The topic to update. Its ``name`` field must be populated.
        :param update_mask: Required. Field mask is used to specify the fields to be overwritten in the Topic
            resource by the update. The fields specified in the update_mask are relative to the resource, not
            the full request. A field will be overwritten if it is in the mask.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        _topic = deepcopy(topic) if isinstance(topic, dict) else Topic.to_dict(topic)
        _topic["name"] = client.topic_path(project_id, location, cluster_id, topic_id)

        result = client.update_topic(
            request={
                "update_mask": update_mask,
                "topic": _topic,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_topic(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        topic_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> None:
        """
        Delete a single topic.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose topic is to be deleted.
        :param topic_id: Required. The ID of the topic to delete.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.topic_path(project_id, location, cluster_id, topic_id)

        client.delete_topic(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def list_consumer_groups(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        page_size: int | None = None,
        page_token: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListConsumerGroupsPager:
        """
        List the consumer groups in a given cluster.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose consumer groups are to be listed.
        :param page_size: Optional. The maximum number of consumer groups to return. The service may return
            fewer than this value. If unset or zero, all consumer groups for the parent is returned.
        :param page_token: Optional. A page token, received from a previous ``ListConsumerGroups`` call.
            Provide this to retrieve the subsequent page. When paginating, all other parameters provided to
            ``ListConsumerGroups`` must match the call that provided the page token.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        parent = client.cluster_path(project_id, location, cluster_id)

        result = client.list_consumer_groups(
            request={
                "parent": parent,
                "page_size": page_size,
                "page_token": page_token,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_consumer_group(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        consumer_group_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.ConsumerGroup:
        """
        Return the properties of a single consumer group.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose consumer group is to be returned.
        :param consumer_group_id: Required. The ID of the consumer group whose configuration to return.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.consumer_group_path(project_id, location, cluster_id, consumer_group_id)

        result = client.get_consumer_group(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_consumer_group(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        consumer_group_id: str,
        consumer_group: types.ConsumerGroup | dict,
        update_mask: FieldMask | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> types.ConsumerGroup:
        """
        Update the properties of a single consumer group.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose topic is to be updated.
        :param consumer_group_id: Required. The ID of the consumer group whose configuration to update.
        :param consumer_group: Required. The consumer_group to update. Its ``name`` field must be populated.
        :param update_mask: Required. Field mask is used to specify the fields to be overwritten in the
            ConsumerGroup resource by the update. The fields specified in the update_mask are relative to the
            resource, not the full request. A field will be overwritten if it is in the mask.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        _consumer_group = (
            deepcopy(consumer_group)
            if isinstance(consumer_group, dict)
            else ConsumerGroup.to_dict(consumer_group)
        )
        _consumer_group["name"] = client.consumer_group_path(
            project_id, location, cluster_id, consumer_group_id
        )

        result = client.update_consumer_group(
            request={
                "update_mask": update_mask,
                "consumer_group": _consumer_group,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_consumer_group(
        self,
        project_id: str,
        location: str,
        cluster_id: str,
        consumer_group_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> None:
        """
        Delete a single consumer group.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud region that the service belongs to.
        :param cluster_id: Required. The ID of the cluster whose consumer group is to be deleted.
        :param consumer_group_id: Required. The ID of the consumer group to delete.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_managed_kafka_client()
        name = client.consumer_group_path(project_id, location, cluster_id, consumer_group_id)

        client.delete_consumer_group(
            request={
                "name": name,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.google.cloud.operators.dataprep import (
    DataprepCopyFlowOperator,
    DataprepDeleteFlowOperator,
    DataprepGetJobGroupOperator,
    DataprepGetJobsForJobGroupOperator,
    DataprepRunFlowOperator,
    DataprepRunJobGroupOperator,
)

GCP_PROJECT_ID = "test-project-id"
DATAPREP_CONN_ID = "dataprep_default"
JOB_ID = 143
FLOW_ID = 128754
NEW_FLOW_ID = 1312
TASK_ID = "dataprep_job"
INCLUDE_DELETED = False
EMBED = ""
DATAPREP_JOB_RECIPE_ID = 1234567
PATH_TO_OUTOUT_FILE = "path_to_output_file"
DATA = {
    "wrangledDataset": {"id": DATAPREP_JOB_RECIPE_ID},
    "overrides": {
        "execution": "dataflow",
        "profiler": False,
        "writesettings": [
            {
                "path": PATH_TO_OUTOUT_FILE,
                "action": "create",
                "format": "csv",
                "compression": "none",
                "header": False,
                "asSingleFile": False,
            }
        ],
    },
}


class TestDataprepGetJobsForJobGroupOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute(self, hook_mock):
        op = DataprepGetJobsForJobGroupOperator(
            dataprep_conn_id=DATAPREP_CONN_ID, job_group_id=JOB_ID, task_id=TASK_ID
        )
        op.execute(context={})
        hook_mock.assert_called_once_with(dataprep_conn_id=DATAPREP_CONN_ID)
        hook_mock.return_value.get_jobs_for_job_group.assert_called_once_with(job_id=JOB_ID)


class TestDataprepGetJobGroupOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute(self, hook_mock):
        op = DataprepGetJobGroupOperator(
            dataprep_conn_id=DATAPREP_CONN_ID,
            project_id=None,
            job_group_id=JOB_ID,
            embed=EMBED,
            include_deleted=INCLUDE_DELETED,
            task_id=TASK_ID,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.get_job_group.assert_called_once_with(
            job_group_id=JOB_ID, embed=EMBED, include_deleted=INCLUDE_DELETED
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.DataprepJobGroupLink")
    @pytest.mark.parametrize(
        ("provide_project_id", "expected_call_count"),
        [
            (True, 1),
            (False, 0),
        ],
    )
    def test_execute_with_project_id_will_persist_link_to_job_group(
        self,
        link_mock,
        _,
        provide_project_id,
        expected_call_count,
    ):
        context = mock.MagicMock()
        project_id = GCP_PROJECT_ID if provide_project_id else None

        op = DataprepGetJobGroupOperator(
            task_id=TASK_ID,
            project_id=project_id,
            dataprep_conn_id=DATAPREP_CONN_ID,
            job_group_id=JOB_ID,
            embed=EMBED,
            include_deleted=INCLUDE_DELETED,
        )
        op.execute(context=context)

        assert link_mock.persist.call_count == expected_call_count
        if provide_project_id:
            link_mock.persist.assert_called_with(
                context=context,
                project_id=project_id,
                job_group_id=JOB_ID,
            )


class TestDataprepRunJobGroupOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute(self, hook_mock):
        op = DataprepRunJobGroupOperator(
            dataprep_conn_id=DATAPREP_CONN_ID,
            body_request=DATA,
            task_id=TASK_ID,
        )
        op.execute(context=None)
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.run_job_group.assert_called_once_with(body_request=DATA)


class TestDataprepCopyFlowOperatorTest:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute_with_default_params(self, hook_mock):
        op = DataprepCopyFlowOperator(
            task_id=TASK_ID,
            dataprep_conn_id=DATAPREP_CONN_ID,
            flow_id=FLOW_ID,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.copy_flow.assert_called_once_with(
            flow_id=FLOW_ID,
            name="",
            description="",
            copy_datasources=False,
        )

    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute_with_specified_params(self, hook_mock):
        op = DataprepCopyFlowOperator(
            task_id=TASK_ID,
            dataprep_conn_id=DATAPREP_CONN_ID,
            flow_id=FLOW_ID,
            name="specified name",
            description="specified description",
            copy_datasources=True,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.copy_flow.assert_called_once_with(
            flow_id=FLOW_ID, name="specified name", description="specified description", copy_datasources=True
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute_with_templated_params(self, _, create_task_instance_of_operator, session):
        dag_id = "test_execute_with_templated_params"
        ti = create_task_instance_of_operator(
            DataprepCopyFlowOperator,
            dag_id=dag_id,
            project_id="{{ dag.dag_id }}",
            task_id=TASK_ID,
            flow_id="{{ dag.dag_id }}",
            name="{{ dag.dag_id }}",
            description="{{ dag.dag_id }}",
        )
        session.add(ti)
        session.commit()
        ti.render_templates()
        assert dag_id == ti.task.project_id
        assert dag_id == ti.task.flow_id
        assert dag_id == ti.task.name
        assert dag_id == ti.task.description

    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.DataprepFlowLink")
    @pytest.mark.parametrize(
        ("provide_project_id", "expected_call_count"),
        [
            (True, 1),
            (False, 0),
        ],
    )
    def test_execute_with_project_id_will_persist_link_to_flow(
        self,
        link_mock,
        hook_mock,
        provide_project_id,
        expected_call_count,
    ):
        hook_mock.return_value.copy_flow.return_value = {"id": NEW_FLOW_ID}
        context = mock.MagicMock()
        project_id = GCP_PROJECT_ID if provide_project_id else None

        op = DataprepCopyFlowOperator(
            task_id=TASK_ID,
            project_id=project_id,
            dataprep_conn_id=DATAPREP_CONN_ID,
            flow_id=FLOW_ID,
            name="specified name",
            description="specified description",
            copy_datasources=True,
        )
        op.execute(context=context)

        assert link_mock.persist.call_count == expected_call_count
        if provide_project_id:
            link_mock.persist.assert_called_with(
                context=context,
                project_id=project_id,
                flow_id=NEW_FLOW_ID,
            )


class TestDataprepDeleteFlowOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute(self, hook_mock):
        op = DataprepDeleteFlowOperator(
            task_id=TASK_ID,
            dataprep_conn_id=DATAPREP_CONN_ID,
            flow_id=FLOW_ID,
        )
        op.execute(context=mock.MagicMock())
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.delete_flow.assert_called_once_with(
            flow_id=FLOW_ID,
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute_with_template_params(self, _, create_task_instance_of_operator, session):
        dag_id = "test_execute_delete_flow_with_template"
        ti = create_task_instance_of_operator(
            DataprepDeleteFlowOperator,
            dag_id=dag_id,
            task_id=TASK_ID,
            flow_id="{{ dag.dag_id }}",
        )
        session.add(ti)
        session.commit()
        ti.render_templates()
        assert dag_id == ti.task.flow_id


class TestDataprepRunFlowOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute(self, hook_mock):
        op = DataprepRunFlowOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT_ID,
            dataprep_conn_id=DATAPREP_CONN_ID,
            flow_id=FLOW_ID,
            body_request={},
        )
        op.execute(context=mock.MagicMock())
        hook_mock.assert_called_once_with(dataprep_conn_id="dataprep_default")
        hook_mock.return_value.run_flow.assert_called_once_with(
            flow_id=FLOW_ID,
            body_request={},
        )

    @pytest.mark.db_test
    @mock.patch("airflow.providers.google.cloud.operators.dataprep.GoogleDataprepHook")
    def test_execute_with_template_params(self, _, create_task_instance_of_operator, session):
        dag_id = "test_execute_run_flow_with_template"
        ti = create_task_instance_of_operator(
            DataprepRunFlowOperator,
            dag_id=dag_id,
            task_id=TASK_ID,
            project_id="{{ dag.dag_id }}",
            flow_id="{{ dag.dag_id }}",
            body_request={},
        )
        session.add(ti)
        session.commit()
        ti.render_templates()

        assert dag_id == ti.task.project_id
        assert dag_id == ti.task.flow_id



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Gen AI operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.google.cloud.hooks.gen_ai import (
    GenAIGenerativeModelHook,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

if TYPE_CHECKING:
    from google.genai.types import (
        ContentListUnion,
        ContentListUnionDict,
        CountTokensConfigOrDict,
        CreateCachedContentConfigOrDict,
        CreateTuningJobConfigOrDict,
        EmbedContentConfigOrDict,
        GenerateContentConfig,
        TuningDatasetOrDict,
    )

    from airflow.providers.common.compat.sdk import Context


class GenAIGenerateEmbeddingsOperator(GoogleCloudBaseOperator):
    """
    Uses the Gemini AI Embeddings API to generate embeddings for words, phrases, sentences, and code.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param model: Required. The name of the model to use for content generation,
        which can be a text-only or multimodal model. For example, `gemini-pro` or
        `gemini-pro-vision`.
    :param contents: Optional. The contents to use for embedding.
    :param config: Optional. Configuration for embeddings.
    :param gcp_conn_id: Optional. The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional. Service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "model", "config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        contents: ContentListUnion | ContentListUnionDict | list[str],
        config: EmbedContentConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.config = config
        self.model = model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Generating text embeddings...")
        response = self.hook.embed_content(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            model=self.model,
            config=self.config,
        )

        self.log.info("Model response: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


class GenAIGenerateContentOperator(GoogleCloudBaseOperator):
    """
    Generate a model response based on given configuration. Input capabilities differ between models, including tuned models.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param model: Required. The name of the model to use for content generation,
        which can be a text-only or multimodal model. For example, `gemini-pro` or
        `gemini-pro-vision`.
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param generation_config: Optional. Generation configuration settings.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "generation_config",
        "location",
        "project_id",
        "impersonation_chain",
        "contents",
        "model",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: ContentListUnionDict,
        model: str,
        generation_config: GenerateContentConfig | dict[str, Any] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.generation_config = generation_config
        self.model = model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.generate_content(
            project_id=self.project_id,
            location=self.location,
            model=self.model,
            contents=self.contents,
            generation_config=self.generation_config,
        )

        self.log.info("Created Content: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


class GenAISupervisedFineTuningTrainOperator(GoogleCloudBaseOperator):
    """
    Create a tuning job to adapt model behavior with a labeled dataset.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param source_model: Required. A pre-trained model optimized for performing natural
        language tasks such as classification, summarization, extraction, content
        creation, and ideation.
    :param training_dataset: Required. Cloud Storage URI of your training dataset. The dataset
        must be formatted as a JSONL file. For best results, provide at least 100 to 500 examples.
    :param tuning_job_config: Optional. Configuration of the Tuning job to be created.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "training_dataset",
        "tuning_job_config",
        "source_model",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        source_model: str,
        training_dataset: TuningDatasetOrDict,
        tuning_job_config: CreateTuningJobConfigOrDict | dict[str, Any] | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.source_model = source_model
        self.training_dataset = training_dataset
        self.tuning_job_config = tuning_job_config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.supervised_fine_tuning_train(
            project_id=self.project_id,
            location=self.location,
            source_model=self.source_model,
            training_dataset=self.training_dataset,
            tuning_job_config=self.tuning_job_config,
        )

        self.log.info("Tuned Model Name: %s", response.tuned_model.model)  # type: ignore[union-attr,arg-type]
        self.log.info("Tuned Model EndpointName: %s", response.tuned_model.endpoint)  # type: ignore[union-attr,arg-type]

        context["ti"].xcom_push(key="tuned_model_name", value=response.tuned_model.model)  # type: ignore[union-attr,arg-type]
        context["ti"].xcom_push(key="tuned_model_endpoint_name", value=response.tuned_model.endpoint)  # type: ignore[union-attr,arg-type]

        result = {
            "tuned_model_name": response.tuned_model.model,  # type: ignore[union-attr,arg-type]
            "tuned_model_endpoint_name": response.tuned_model.endpoint,  # type: ignore[union-attr,arg-type]
        }

        return result


class GenAICountTokensOperator(GoogleCloudBaseOperator):
    """
    Use Count Tokens API to calculate the number of input tokens before sending a request to Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param model: Required. Model, supporting prompts with text-only input,
        including natural language tasks, multi-turn text and code chat,
        and code generation. It can output text and code.
    :param config: Optional. Configuration for Count Tokens.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "model", "config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: ContentListUnion | ContentListUnionDict,
        model: str,
        config: CountTokensConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.model = model
        self.config = config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.count_tokens(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            model=self.model,
            config=self.config,
        )

        self.log.info("Total tokens: %s", response.total_tokens)
        context["ti"].xcom_push(key="total_tokens", value=response.total_tokens)


class GenAICreateCachedContentOperator(GoogleCloudBaseOperator):
    """
    Create CachedContent resource to reduce the cost of requests that contain repeat content with high input token counts.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model: Required. The name of the publisher model to use for cached content.
    :param cached_content_config: Optional. Configuration of the Cached Content.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "model", "cached_content_config")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model: str,
        cached_content_config: CreateCachedContentConfigOrDict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.model = model
        self.cached_content_config = cached_content_config
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenAIGenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        cached_content_name = self.hook.create_cached_content(
            project_id=self.project_id,
            location=self.location,
            model=self.model,
            cached_content_config=self.cached_content_config,
        )

        self.log.info("Cached Content Name: %s", cached_content_name)
        context["ti"].xcom_push(key="cached_content", value=cached_content_name)

        return cached_content_name



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime, timezone
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import (
    ACCESS_KEY_ID,
    AWS_ACCESS_KEY,
    AWS_S3_DATA_SOURCE,
    AWS_SECRET_ACCESS_KEY,
    BUCKET_NAME,
    GCS_DATA_SINK,
    INCLUDE_PREFIXES,
    OBJECT_CONDITIONS,
    OVERWRITE_OBJECTS_ALREADY_EXISTING_IN_SINK,
    PATH,
    PROJECT_ID,
    SCHEDULE,
    SCHEDULE_END_DATE,
    SCHEDULE_START_DATE,
    STATUS,
    TRANSFER_OPTIONS,
    TRANSFER_SPEC,
    CloudDataTransferServiceHook,
    GcpTransferJobsStatus,
)
from airflow.providers.google.cloud.hooks.gcs import GCSHook, _parse_gcs_url, gcs_object_is_directory
from airflow.providers.google.cloud.triggers.cloud_storage_transfer_service import (
    CloudStorageTransferServiceCreateJobsTrigger,
)

try:
    from airflow.providers.amazon.aws.operators.s3 import S3ListOperator
except ImportError:
    from airflow.providers.amazon.aws.operators.s3_list import S3ListOperator  # type: ignore[no-redef]

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class S3ToGCSOperator(S3ListOperator):
    """
    Synchronizes an S3 key, possibly a prefix, with a Google Cloud Storage destination path.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:S3ToGCSOperator`

    :param bucket: The S3 bucket where to find the objects. (templated)
    :param prefix: Prefix string which filters objects whose name begin with
        such prefix. (templated)
    :param apply_gcs_prefix: (Optional) Whether to replace source objects' path by given GCS destination path.
        If apply_gcs_prefix is False (default), then objects from S3 will be copied to GCS bucket into a given
        GSC path and the source path will be place inside. For example,
        <s3_bucket><s3_prefix><content> => <gcs_prefix><s3_prefix><content>

        If apply_gcs_prefix is True, then objects from S3 will be copied to GCS bucket into a given
        GCS path and the source path will be omitted. For example:
        <s3_bucket><s3_prefix><content> => <gcs_prefix><content>

    :param delimiter: the delimiter marks key hierarchy. (templated)
    :param aws_conn_id: The source S3 connection
    :param verify: Whether or not to verify SSL certificates for S3 connection.
        By default SSL certificates are verified.
        You can provide the following values:

        - ``False``: do not validate SSL certificates. SSL will still be used
                 (unless use_ssl is False), but SSL certificates will not be
                 verified.
        - ``path/to/cert/bundle.pem``: A filename of the CA cert bundle to uses.
                 You can specify this argument if you want to use a different
                 CA cert bundle than the one used by botocore.
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param dest_gcs: The destination Google Cloud Storage bucket and prefix
        where you want to store the files. (templated)
    :param replace: Whether you want to replace existing destination files
        or not.
    :param gzip: Option to compress file for upload. Parameter ignored in deferrable mode.
    :param google_impersonation_chain: Optional Google service account to impersonate using
        short-term credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: Run operator in the deferrable mode
    :param poll_interval: time in seconds between polling for job completion.
        The value is considered only when running in deferrable mode. Must be greater than 0.

    **Example**:

    .. code-block:: python

       s3_to_gcs_op = S3ToGCSOperator(
           task_id="s3_to_gcs_example",
           bucket="my-s3-bucket",
           prefix="data/customers-201804",
           gcp_conn_id="google_cloud_default",
           dest_gcs="gs://my.gcs.bucket/some/customers/",
           replace=False,
           gzip=True,
           dag=my_dag,
       )

    Note that ``bucket``, ``prefix``, ``delimiter`` and ``dest_gcs`` are
    templated, so you can use variables in them if you wish.
    """

    template_fields: Sequence[str] = (
        "bucket",
        "prefix",
        "delimiter",
        "dest_gcs",
        "google_impersonation_chain",
    )
    ui_color = "#e09411"
    transfer_job_max_files_number = 1000

    def __init__(
        self,
        *,
        bucket,
        prefix="",
        apply_gcs_prefix=False,
        delimiter="",
        aws_conn_id="aws_default",
        verify=None,
        gcp_conn_id="google_cloud_default",
        dest_gcs=None,
        replace=False,
        gzip=False,
        google_impersonation_chain: str | Sequence[str] | None = None,
        deferrable=conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ):
        super().__init__(bucket=bucket, prefix=prefix, delimiter=delimiter, aws_conn_id=aws_conn_id, **kwargs)
        self.apply_gcs_prefix = apply_gcs_prefix
        self.gcp_conn_id = gcp_conn_id
        self.dest_gcs = dest_gcs
        self.replace = replace
        self.verify = verify
        self.gzip = gzip
        self.google_impersonation_chain = google_impersonation_chain
        self.deferrable = deferrable
        if poll_interval <= 0:
            raise ValueError("Invalid value for poll_interval. Expected value greater than 0")
        self.poll_interval = poll_interval

    def _check_inputs(self) -> None:
        if self.dest_gcs and not gcs_object_is_directory(self.dest_gcs):
            self.log.info(
                "Destination Google Cloud Storage path is not a valid "
                '"directory", define a path that ends with a slash "/" or '
                "leave it empty for the root of the bucket."
            )
            raise AirflowException(
                'The destination Google Cloud Storage path must end with a slash "/" or be empty.'
            )

    def _get_files(self, context: Context, gcs_hook: GCSHook) -> list[str]:
        # use the super method to list all the files in an S3 bucket/key
        s3_objects = super().execute(context)

        if not self.replace:
            s3_objects = self.exclude_existing_objects(s3_objects=s3_objects, gcs_hook=gcs_hook)

        return s3_objects

    def execute(self, context: Context):
        self._check_inputs()
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.google_impersonation_chain,
        )
        s3_hook = S3Hook(aws_conn_id=self.aws_conn_id, verify=self.verify)

        s3_objects = self._get_files(context, gcs_hook)
        if not s3_objects:
            self.log.info("In sync, no files needed to be uploaded to Google Cloud Storage")

        elif self.deferrable:
            self.transfer_files_async(s3_objects, gcs_hook, s3_hook)
        else:
            self.transfer_files(s3_objects, gcs_hook, s3_hook)

        return s3_objects

    def exclude_existing_objects(self, s3_objects: list[str], gcs_hook: GCSHook) -> list[str]:
        """Excludes from the list objects that already exist in GCS bucket."""
        bucket_name, object_prefix = _parse_gcs_url(self.dest_gcs)

        existing_gcs_objects = set(gcs_hook.list(bucket_name, prefix=object_prefix))

        s3_paths = set(self.gcs_to_s3_object(gcs_object=gcs_object) for gcs_object in existing_gcs_objects)
        s3_objects_reduced = list(set(s3_objects) - s3_paths)

        if s3_objects_reduced:
            self.log.info("%s files are going to be synced: %s.", len(s3_objects_reduced), s3_objects_reduced)
        else:
            self.log.info("There are no new files to sync. Have a nice day!")
        return s3_objects_reduced

    def s3_to_gcs_object(self, s3_object: str) -> str:
        """
        Transform S3 path to GCS path according to the operator's logic.

        If apply_gcs_prefix == True then <s3_prefix><content> => <gcs_prefix><content>
        If apply_gcs_prefix == False then <s3_prefix><content> => <gcs_prefix><s3_prefix><content>

        """
        gcs_bucket, gcs_prefix = _parse_gcs_url(self.dest_gcs)
        if self.apply_gcs_prefix:
            gcs_object = s3_object.replace(self.prefix, gcs_prefix, 1)
            return gcs_object
        return gcs_prefix + s3_object

    def gcs_to_s3_object(self, gcs_object: str) -> str:
        """
        Transform GCS path to S3 path according to the operator's logic.

        If apply_gcs_prefix == True then <gcs_prefix><content> => <s3_prefix><content>
        If apply_gcs_prefix == False then <gcs_prefix><s3_prefix><content> => <s3_prefix><content>

        """
        gcs_bucket, gcs_prefix = _parse_gcs_url(self.dest_gcs)
        s3_object = gcs_object.replace(gcs_prefix, "", 1)
        if self.apply_gcs_prefix:
            return self.prefix + s3_object
        return s3_object

    def transfer_files(self, s3_objects: list[str], gcs_hook: GCSHook, s3_hook: S3Hook) -> None:
        if s3_objects:
            dest_gcs_bucket, dest_gcs_object_prefix = _parse_gcs_url(self.dest_gcs)
            for obj in s3_objects:
                # GCS hook builds its own in-memory file, so we have to create
                # and pass the path
                file_object = s3_hook.get_key(obj, self.bucket)
                with NamedTemporaryFile(mode="wb", delete=True) as file:
                    file_object.download_fileobj(file)
                    file.flush()
                    gcs_file = self.s3_to_gcs_object(s3_object=obj)
                    gcs_hook.upload(dest_gcs_bucket, gcs_file, file.name, gzip=self.gzip)

            self.log.info("All done, uploaded %d files to Google Cloud Storage", len(s3_objects))

    def transfer_files_async(self, files: list[str], gcs_hook: GCSHook, s3_hook: S3Hook) -> None:
        """Submit Google Cloud Storage Transfer Service job to copy files from AWS S3 to GCS."""
        if not len(files):
            raise ValueError("List of transferring files cannot be empty")
        job_names = self.submit_transfer_jobs(files=files, gcs_hook=gcs_hook, s3_hook=s3_hook)

        self.defer(
            trigger=CloudStorageTransferServiceCreateJobsTrigger(
                project_id=gcs_hook.project_id,
                gcp_conn_id=self.gcp_conn_id,
                job_names=job_names,
                poll_interval=self.poll_interval,
            ),
            method_name="execute_complete",
        )

    def submit_transfer_jobs(self, files: list[str], gcs_hook: GCSHook, s3_hook: S3Hook) -> list[str]:
        now = datetime.now(tz=timezone.utc)
        one_time_schedule = {"day": now.day, "month": now.month, "year": now.year}

        gcs_bucket, gcs_prefix = _parse_gcs_url(self.dest_gcs)
        config = s3_hook.conn_config

        body: dict[str, Any] = {
            PROJECT_ID: gcs_hook.project_id,
            STATUS: GcpTransferJobsStatus.ENABLED,
            SCHEDULE: {
                SCHEDULE_START_DATE: one_time_schedule,
                SCHEDULE_END_DATE: one_time_schedule,
            },
            TRANSFER_SPEC: {
                AWS_S3_DATA_SOURCE: {
                    BUCKET_NAME: self.bucket,
                    AWS_ACCESS_KEY: {
                        ACCESS_KEY_ID: config.aws_access_key_id,
                        AWS_SECRET_ACCESS_KEY: config.aws_secret_access_key,
                    },
                },
                OBJECT_CONDITIONS: {
                    INCLUDE_PREFIXES: [],
                },
                GCS_DATA_SINK: {BUCKET_NAME: gcs_bucket, PATH: gcs_prefix},
                TRANSFER_OPTIONS: {
                    OVERWRITE_OBJECTS_ALREADY_EXISTING_IN_SINK: self.replace,
                },
            },
        }

        # max size of the field 'transfer_job.transfer_spec.object_conditions.include_prefixes' is 1000,
        # that's why we submit multiple jobs transferring 1000 files each.
        # See documentation below
        # https://cloud.google.com/storage-transfer/docs/reference/rest/v1/TransferSpec#ObjectConditions
        chunk_size = self.transfer_job_max_files_number
        job_names = []
        transfer_hook = self.get_transfer_hook()
        for i in range(0, len(files), chunk_size):
            files_chunk = files[i : i + chunk_size]
            body[TRANSFER_SPEC][OBJECT_CONDITIONS][INCLUDE_PREFIXES] = files_chunk
            job = transfer_hook.create_transfer_job(body=body)

            self.log.info("Submitted job %s to transfer %s file(s).", job["name"], len(files_chunk))
            job_names.append(job["name"])

        if len(files) > chunk_size:
            self.log.info("Overall submitted %s job(s) to transfer %s file(s).", len(job_names), len(files))

        return job_names

    def execute_complete(self, context: Context, event: dict[str, Any]) -> None:
        """
        Return immediately and relies on trigger to throw a success event. Callback for the trigger.

        Relies on trigger to throw an exception, otherwise it assumes execution was
        successful.
        """
        if event["status"] == "error":
            raise AirflowException(event["message"])
        self.log.info("%s completed with response %s ", self.task_id, event["message"])

    def get_transfer_hook(self):
        return CloudDataTransferServiceHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.google_impersonation_chain,
        )

    def get_openlineage_facets_on_start(self):
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.google.cloud.hooks.gcs import _parse_gcs_url
        from airflow.providers.google.cloud.openlineage.utils import extract_ds_name_from_gcs_path
        from airflow.providers.openlineage.extractors import OperatorLineage

        gcs_bucket, gcs_blob = _parse_gcs_url(self.dest_gcs)
        if not self.apply_gcs_prefix:
            gcs_blob += self.prefix

        return OperatorLineage(
            inputs=[Dataset(namespace=f"s3://{self.bucket}", name=self.prefix.strip("/") or "/")],
            outputs=[Dataset(namespace=f"gs://{gcs_bucket}", name=extract_ds_name_from_gcs_path(gcs_blob))],
        )
